import datetime
from time import sleep
from multiprocessing import Process
import urllib.request

from .exceptions import RunError, LtemetrError
from .logging import get_logger
from .lteinfo import measure_lte

CSV_HEADER = "Time,Lat,Long,RSRP,SINR Rx[0],RSRQ,RSSI,IP Thrpt DL,MCC,MNC,LAC,"\
             "Cell Id,PCI,DL EARFCN,Bandwidth,QPSK Rate,16-QAM Rate,64-QAM Rate,"\
             "256-QAM Rate,Carrier Aggregation DL"

logger = get_logger()


class MeasurementTypes:
    TRAIN = "TRAIN"
    CAR = "CAR"


def wait_to_full_sec():
    sleep(1 - datetime.datetime.utcnow().microsecond / 1000000)


def log(time, lte, dwl):
    pattern = "{:%d.%m.%Y %H:%M:%S},{},{},{},{},{},{},{:.2f},{},{},{},{},{},{},{}"
    return pattern.format(
        time,
        lte.get("lat", ""),
        lte.get("lon", ""),
        lte.get("RSRP", ""),
        lte.get("SINR Rx[0]", ""),
        lte.get("RSRQ", ""),
        lte.get("RSSI", ""),
        dwl,
        lte.get("MCC", ""),
        lte.get("MNC", ""),
        lte.get("LAC", ""),
        lte.get("Cell Id", ""),
        lte.get("PCI", ""),
        lte.get("DL EARFCN", ""),
        lte.get("Bandwidth", "")
    )


class Measurement:
    def __init__(self, download_src=None, iface=None, modem_console_path=None,
                 stop_event=lambda: True):
        self.stop_event = stop_event
        self.log = [CSV_HEADER]
        self.last_kbps = 0
        self.last_lte = {}
        self.last_time = None
        self._download_src = download_src
        self._iface = iface
        self._modem_console_path = modem_console_path

    def _get_rx_bytes(self):
        try:
            with open("/sys/class/net/{}/statistics/rx_bytes".format(self._iface), "r") as f:
                return int(f.readline())
        except OSError as e:
            raise RunError("Measuring on non-existing interface ({})".format(e)) from e
        except ValueError as e:
            raise RunError("rx_bytes must be an integer ({})".format(e)) from e

    def _download_runner(self):
        logger.debug("Spawning download runner to get {}".format(self._download_src))
        try:
            urllib.request.urlretrieve(self._download_src, "/dev/null")
        except Exception as e:
            logger.error("Download runner error: %s", e)

    def _cleanup(self, proc):
        if proc.is_alive():
            proc.terminate()

    def measure(self):
        proc = Process(target=self._download_runner)
        sleep(1)
        proc.start()

        # echo "Time,Lat,Long,RSRP,SINR Rx[0],RSRQ,RSSI,IP Thrpt DL,MCC,MNC,LAC,\
        # Cell Id,PCI,DL EARFCN,Bandwidth,QPSK Rate,16-QAM Rate,64-QAM Rate,\
        # 256-QAM Rate,Carrier Aggregation DL"

        try:
            wait_to_full_sec()
            last_time = datetime.datetime.utcnow()
            last_count = self._get_rx_bytes()
            while proc.is_alive() and not self.stop_event():
                wait_to_full_sec()
                count = self._get_rx_bytes()
                now = datetime.datetime.utcnow()
                KBPS = (count - last_count) * 8 / 1000 / (now.timestamp() - last_time.timestamp())
                lte_data = measure_lte(self._modem_console_path)
                logger.debug(log(now, lte_data, KBPS))
                self.log.append(log(now, lte_data, KBPS))

                self.last_kbps = KBPS
                self.last_lte = lte_data
                self.last_time = now

                last_count = count
                last_time = now
            logger.debug("Ending measurement")
        except Exception as e:
            self._cleanup(proc)
            raise LtemetrError(str(e)) from e

        self._cleanup(proc)
